/**
 * <p>Title: Haplotype</p>
 * <p>Description: Approximation Algorithms for Robust tag SNPs</p>
 * <p>Copyright: Copyright (c) 2004</p>
 * <p>Company: CSIE.NTU</p>
 * @author Yao-Ting Huang
 * @version 1.0
 */

import java.util.*;
import java.io.*;

public class Greedy1 {
  Vector reduction;
  int NumberOfMissingData, size, total=0;
  BufferedWriter bw, out;
  String BlockID;

  public Greedy1(int NumberOfMissingData) {
    this.reduction = new Vector();
    this.NumberOfMissingData = NumberOfMissingData;
    this.init();
  }

  private void init(){
   try{
    this.bw = new BufferedWriter(new FileWriter(new File("Reduction1.txt")));
    this.out = new BufferedWriter(new FileWriter(new File("Greedy1_"+this.NumberOfMissingData+".txt")));
   }
   catch(IOException ee){
    ee.printStackTrace();
     }
  }


  public void Reduce(Vector vt, String BlockID) {
    StringBuffer sb = (StringBuffer) vt.elementAt(0);
    this.BlockID = BlockID;

    this.size = vt.size()*(vt.size()-1)/2;

    //reset reduction vector;
    reduction.clear();
    for (int i = 0; i < sb.length(); i++) {
      Bipartite gt = new Bipartite();
      gt.locus = i;
      gt.edges = new int[this.size];
      int index = 0;
      for (int j = 0; j < vt.size(); j++) {
        StringBuffer sb1 = (StringBuffer) vt.elementAt(j);
        for (int k = j + 1; k < vt.size(); k++) {
          StringBuffer sb2 = (StringBuffer) vt.elementAt(k);
          if ((sb1.charAt(i) != sb2.charAt(i)) && sb1.charAt(i)!='n' && sb2.charAt(i)!='n') {
          //if ((sb1.charAt(i) != sb2.charAt(i))) {
            gt.edges[gt.count] = index;
            gt.count++;
          }
          index++;
        }
      }
      //store the transformed result
      reduction.addElement(gt);
    }

    //write reduction result
    try{
      for (int l = 0; l < reduction.size(); l++) {
        Bipartite temp = (Bipartite) reduction.elementAt(l);
        bw.write(BlockID+"\t"+temp.count);
        for(int m=0;m<temp.edges.length;m++){
          bw.write("\t"+temp.edges[m]);
        }
        bw.newLine();
        bw.flush();
      }
    }
    catch (Exception ee) {
      ee.printStackTrace();
    }

  }

  public void run(){
   int[][] table = new int[this.NumberOfMissingData + 1][this.size];
   int NumberOfTagSNP = 0;
   int[] tablesize = new int[this.NumberOfMissingData + 1];
   Vector result = new Vector();

   //initialize
   for(int m=0;m<tablesize.length;m++){
     tablesize[m] = this.size;
     //Set the table with depth 1
     for (int i = 0; i < table[m].length; i++)
       table[m][i] = 1;
   }


   //start the first greedy algorithm
   for(int j = 0; j < (this.NumberOfMissingData + 1); j++){
    while(tablesize[j] > 0){
      //Greedy stage; pick one set that maximizes covered elements
      Bipartite bp = findMax(table[j]);
      if (bp == null) {
        return;
      }

      result.addElement(bp);

      int count = bp.count;
      while (count > 0) {
        for (int k = 0; k < bp.count; k++) {
          //this column is not finished
          count--;
          if (table[j][bp.edges[k]] > 0) {
            table[j][bp.edges[k]]--;
            tablesize[j]--;
          }
          else {
            //this row is finished; check following rows
            for(int x=j+1; x<this.NumberOfMissingData+1;x++){
              if(table[x][bp.edges[k]] > 0){
                table[x][bp.edges[k]]--;
                tablesize[x]--;
                break;
              }//end of if
            }//end of inner for
          }//end of else
        } //end of inner for loop
      } //end of inner while loop
    } //end of outer while loop
  } //end of outer for loop

   this.total = this.total + result.size();

   //Write result
   try {
     if(result.size() > 0){
       this.out.write(this.BlockID + "\t" + result.size());
       for (int l = 0; l < result.size(); l++) {
         Bipartite temp = (Bipartite) result.elementAt(l);
         this.out.write("\t" + temp.locus);
       }
       this.out.newLine();
       this.out.flush();
     }
 }//end of try
   catch(IOException ee){
    ee.printStackTrace();
  }

}

  private Bipartite findMax(int[] table){
   Bipartite candidate = null;
   int max = 0;
   int index = 0;
   boolean found = false;

   for(int i=0; i<reduction.size(); i++){
    Bipartite temp = (Bipartite) reduction.elementAt(i);
    int count = 0;

    for(int j = 0; j < temp.count; j++){
     //this column is not finished
     if(table[temp.edges[j]]>0 ){
       count++;
       found = true;
     }
    }//end of for loop

    if(count>max){
     max = count;
     index = i;
    }//end of if
   }//end of for loop

   //remove the chosen SNP
   if(found)
     return (Bipartite) reduction.remove(index);
   else return null;
  }

  public void close(){
   try{
     this.out.write("Total SNPs: "+this.total);
     this.out.flush();
    this.bw.close();
    this.out.close();
   }
   catch(IOException ee){
    ee.printStackTrace();
     }

  }

  /**
   * <p>Title: Haplotype</p>
   *
   * <p>Description: Approximation Algorithms for Robust tag SNPs</p>
   *
   * <p>Copyright: Copyright (c) 2004</p>
   *
   * <p>Company: CSIE.NTU</p>
   * @author Yao-Ting Huang
   * @version 1.0
   */
  private class Bipartite implements Serializable{
    /**
     * Bipartite
     *
     * @param length int
     */

    int count = 0; //number of edges
    int locus;  //locus of the SNP
    int[] edges;
  }
}
